
from ..handlers.base import Filter, LogRecord

class LevelFilter(Filter):
    """filter that filters records based on log level."""

    def __init__(self, min_level: str = "TRACE"):
        from dlogger.logger import dLogger
        level_data = dLogger.LEVELS.get(min_level.upper())
        self._min_level = level_data[0] if level_data else 10

    def filter(self, record: LogRecord) -> bool:
        """return True if record level is >= min_level."""
        return record.level_value >= self._min_level




from ..handlers.base import Filter, LogRecord

class KeywordFilter(Filter):
    """filter that excludes records containing specific keywords."""

    def __init__(self, exclude: list[str], case_sensitive: bool = False):
        self.exclude = exclude
        self.case_sensitive = case_sensitive

    def filter(self, record: LogRecord) -> bool:
        msg = record.message if self.case_sensitive else record.message.lower()
        keywords = self.exclude if self.case_sensitive else [k.lower() for k in self.exclude]
        return not any(kw in msg for kw in keywords)




from datetime import datetime
from typing import Optional, Literal, List, Union
import threading
import inspect
import os

from .handlers.base import Handler, LogRecord
from .handlers.console import ConsoleHandler
from .handlers.file import FileHandler
from .formatters.exception import ExceptionFormatter

class dLogger:
    """main logger class - facade over handlers."""

    LEVELS = {
        "TRACE": (10, "#00bcd4"),
        "DEBUG": (10, "#3b82f6"),
        "INFO": (20, "#ffffff"),
        "SUCCESS": (30, "#4caf50"),
        "WARNING": (30, "#ff9800"),
        "ERROR": (40, "#f44336"),
        "CRITICAL": (50, "#f44336"),
    }

    def __init__(self):
        self._level = 10
        self._handlers: List[Handler] = []
        self._lock = threading.Lock()
        self._context_cache = {}

        self.add_handler(ConsoleHandler(level="TRACE"))

    @property
    def handlers(self) -> List[Handler]:
        return self._handlers

    def add_handler(self, handler: Handler):
        """add a handler to the logger."""
        with self._lock:
            self._handlers.append(handler)

    def remove_handler(self, handler: Handler):
        """remove a handler from the logger."""
        with self._lock:
            if handler in self._handlers:
                self._handlers.remove(handler)

    def configure(
        self,
        level: Literal["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"] = "DEBUG",
        log_file: Optional[str] = None,
        show_path: bool = True,
        rotation: Optional[str] = None,
        retention: Optional[str] = None,
        compression: bool = False,
        time_format: Literal[
            "%Y-%m-%d %H:%M:%S",
            "%H:%M:%S",
            "%d.%m.%Y %H:%M",
            "%Y-%m-%dT%H:%M:%S",
            "%d/%m/%Y %H:%M:%S",
            "%Y-%m-%d %H:%M:%S.%f"
        ] = "%Y-%m-%d %H:%M:%S"
    ):
        """
        configure logger settings.

        args:
            level: Logging level
            log_file: Path to log file
            show_path: Show module:function: in logs
            rotation: Log rotation ("10MB", "1GB", "1 day", "12 hours")
            retention: How long to keep logs ("7 days", "1 month")
            compression: Compress old logs to .gz
            time_format: Time format string
        """
        self._level = self.LEVELS.get(level.upper(), (10,))[0]

        for handler in self._handlers:
            handler.set_level(level)
            if isinstance(handler, ConsoleHandler):
                handler._show_path = show_path

        if log_file:
            file_handler = FileHandler(
                filename=log_file,
                level=level,
                rotation=rotation,
                retention=retention,
                compression=compression,
                time_format=time_format,
            )
            self.add_handler(file_handler)

        return self

    def _get_context(self) -> str:
        frame = inspect.currentframe()
        try:
            caller_frame = frame.f_back
            if not caller_frame:
                return "unknown"
            
            caller_frame = caller_frame.f_back
            if not caller_frame:
                return "unknown"
            
            caller_frame = caller_frame.f_back
            if not caller_frame:
                return "unknown"
            
            filename = caller_frame.f_code.co_filename
            lineno = caller_frame.f_lineno
            
            cache_key = (filename, lineno)
            if cache_key in self._context_cache:
                return self._context_cache[cache_key]
            
            if "dlogger" not in caller_frame.f_globals.get("__name__", ""):
                module = caller_frame.f_globals.get("__name__", "unknown")
                function = caller_frame.f_code.co_name
                result = f"{module}:{function}:"
                
                with self._lock:
                    if len(self._context_cache) < 128:
                        self._context_cache[cache_key] = result
                
                return result
            
            return "unknown"
        finally:
            del frame

    def _log(self, level_name: str, msg: str):
        level_data = self.LEVELS.get(level_name)
        if not level_data:
            return

        level_val, clr = level_data
        if level_val < self._level:
            return

        context = self._get_context()
        now = datetime.now()

        record = LogRecord(
            level=level_name,
            level_value=level_val,
            message=msg,
            context=context,
            timestamp=now,
            color=clr,
        )

        with self._lock:
            for handler in self._handlers:
                handler.emit(record)

    def trace(self, msg: str):
        self._log("TRACE", msg)

    def debug(self, msg: str):
        self._log("DEBUG", msg)

    def info(self, msg: str):
        self._log("INFO", msg)

    def success(self, msg: str):
        self._log("SUCCESS", msg)

    def warning(self, msg: str):
        self._log("WARNING", msg)

    def error(self, msg: str):
        self._log("ERROR", msg)

    def critical(self, msg: str):
        self._log("CRITICAL", msg)

    def exception(self, msg: str, exc: Optional[BaseException] = None):
        """log exception with traceback.
        
        args:
            msg: message
            exc: exception object (optional, uses sys.exc_info() if not provided)
        """
        if exc is None:
            exc = ExceptionFormatter.get_current_exception()
        
        if exc:
            tb = ExceptionFormatter.format_exception(exc)
            full_msg = f"{msg}\n{tb}"
        else:
            full_msg = msg
        
        self._log("ERROR", full_msg)

logger = dLogger()




from typing import Optional
from dcolor import color

from .base import Handler, LogRecord

class ConsoleHandler(Handler):
    """handler that writes log records to console/stdout."""

    def __init__(
        self,
        level: str = "TRACE",
        formatter: Optional[object] = None,
        show_path: bool = True,
    ):
        super().__init__(level=level, formatter=formatter)
        self._show_path = show_path
        self._separator = color('|', 'white')
        self._dash = color('-', 'white')

    def emit(self, record: LogRecord):
        """emit a log record to console."""
        if not self._should_log(record):
            return

        time_str = record.timestamp.strftime("%Y-%m-%d %H:%M:%S")
        context = record.context if self._show_path else ""

        is_critical = record.level == "CRITICAL"

        msg = (
            f"{color(time_str, '#4caf50')} "
            f"{self._separator} {color(f'{record.level: <8}', record.color or '#ffffff', 'bold', *(('underline',) if is_critical else ()))} "
            f"{self._separator} {color(context, '#00bcd4')} "
            f"{self._dash} {record.message}"
        )
        print(msg)




from abc import ABC, abstractmethod
from typing import Optional, List, Any
from datetime import datetime

class Formatter:
    """base formatter interface."""

    def format(self, record: "LogRecord") -> str:
        """format a log record."""
        raise NotImplementedError

class LogRecord:
    """represents a single log event."""

    def __init__(
        self,
        level: str,
        level_value: int,
        message: str,
        context: str,
        timestamp: datetime,
        color: Optional[str] = None,
    ):
        self.level = level
        self.level_value = level_value
        self.message = message
        self.context = context
        self.timestamp = timestamp
        self.color = color

class Filter(ABC):
    """base filter interface."""

    @abstractmethod
    def filter(self, record: LogRecord) -> bool:
        """return True if record should be logged."""
        raise NotImplementedError

class Handler(ABC):
    """abstract base class for log handlers."""

    def __init__(
        self,
        level: str = "TRACE",
        formatter: Optional[Formatter] = None,
        filters: Optional[List[Filter]] = None,
    ):
        self._level = 10
        self._formatter = formatter
        self._filters = filters or []
        self.set_level(level)

    @property
    def level(self) -> int:
        return self._level

    @property
    def formatter(self) -> Optional[Formatter]:
        return self._formatter

    @formatter.setter
    def formatter(self, value: Optional[Formatter]):
        self._formatter = value

    @property
    def filters(self) -> List[Filter]:
        return self._filters

    def add_filter(self, filter_obj: Filter):
        """add a filter to this handler."""
        self._filters.append(filter_obj)

    def set_level(self, level: str):
        """set the handler's minimum log level."""
        from dlogger.logger import dLogger
        level_data = dLogger.LEVELS.get(level.upper())
        if level_data:
            self._level = level_data[0]

    def _should_log(self, record: LogRecord) -> bool:
        """check if record should be logged based on level and filters."""
        if record.level_value < self._level:
            return False

        for filter_obj in self._filters:
            if not filter_obj.filter(record):
                return False
        return True

    @abstractmethod
    def emit(self, record: LogRecord):
        """emit a log record. Must be implemented by subclasses."""
        raise NotImplementedError

    def close(self):
        """close the handler and release resources."""
        pass




from typing import Optional
from datetime import datetime, timedelta
from typing import Optional, Literal
import threading
import atexit
import gzip
import os

from .base import Handler, LogRecord

class FileHandler(Handler):
    """handler that writes log records to a file with rotation and compression support."""

    def __init__(
        self,
        filename: str,
        level: str = "TRACE",
        formatter: Optional[object] = None,
        rotation: Optional[str] = None,
        retention: Optional[str] = None,
        compression: bool = False,
        buffer_size: int = 100,
        time_format: str = "%Y-%m-%d %H:%M:%S",
    ):
        super().__init__(level=level, formatter=formatter)
        self._filename = filename
        self._rotation_size = None
        self._rotation_time = None
        self._retention_days = None
        self._compression = compression
        self._current_file_creation = None
        self._time_format = time_format

        self._lock = threading.Lock()
        self._buffer = []
        self._buffer_size = buffer_size
        self._log_count = 0
        self._check_rotation_every = 100

        self._ensure_log_directory()

        if rotation:
            self._parse_rotation(rotation)

        if retention:
            self._retention_days = self._parse_retention(retention)

        if self._retention_days:
            self._cleanup_old_logs()

        atexit.register(self.close)

    def _parse_rotation(self, rotation: str):
        rotation = rotation.strip().lower()

        if "kb" in rotation or "mb" in rotation or "gb" in rotation:
            value = float(rotation.replace("kb", "").replace("mb", "").replace("gb", "").strip())

            if "kb" in rotation:
                self._rotation_size = int(value * 1024)
            elif "mb" in rotation:
                self._rotation_size = int(value * 1024 * 1024)
            elif "gb" in rotation:
                self._rotation_size = int(value * 1024 * 1024 * 1024)

        elif "hour" in rotation or "day" in rotation or "week" in rotation:
            parts = rotation.split()
            value = int(parts[0])
            unit = parts[1] if len(parts) > 1 else rotation.replace(str(value), "").strip()

            if "hour" in unit:
                self._rotation_time = value * 3600
            elif "day" in unit:
                self._rotation_time = value * 86400
            elif "week" in unit:
                self._rotation_time = value * 604800

    def _parse_retention(self, retention: str) -> int:
        retention = retention.strip().lower()
        parts = retention.split()
        value = int(parts[0])
        unit = parts[1] if len(parts) > 1 else retention.replace(str(value), "").strip()

        if "day" in unit:
            return value
        elif "week" in unit:
            return value * 7
        elif "month" in unit:
            return value * 30
        return value

    def _ensure_log_directory(self):
        log_dir = os.path.dirname(self._filename)
        if log_dir and not os.path.exists(log_dir):
            os.makedirs(log_dir, exist_ok=True)

    def _should_rotate(self) -> bool:
        if not os.path.exists(self._filename):
            return False

        if self._rotation_size:
            file_size = os.path.getsize(self._filename)
            if file_size >= self._rotation_size:
                return True

        if self._rotation_time:
            if self._current_file_creation is None:
                self._current_file_creation = datetime.now()

            time_diff = (datetime.now() - self._current_file_creation).total_seconds()
            if time_diff >= self._rotation_time:
                return True
        return False

    def _rotate_log(self):
        if not os.path.exists(self._filename):
            return

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        rotated_name = f"{self._filename}.{timestamp}"

        try:
            os.rename(self._filename, rotated_name)

            if self._compression:
                self._compress_file(rotated_name)

            self._current_file_creation = datetime.now()

            if self._retention_days:
                self._cleanup_old_logs()
        except Exception as e:
            print(f"⚠️ Error during log rotation: {e}")

    def _compress_file(self, filepath: str):
        try:
            with open(filepath, 'rb') as f_in:
                with gzip.open(f"{filepath}.gz", 'wb') as f_out:
                    f_out.writelines(f_in)
            os.remove(filepath)
        except Exception as e:
            print(f"⚠️ Error during file compression: {e}")

    def _cleanup_old_logs(self):
        if not self._retention_days:
            return

        try:
            log_dir = os.path.dirname(self._filename) or "."
            base_name = os.path.basename(self._filename)

            cutoff_date = datetime.now() - timedelta(days=self._retention_days)

            for filename in os.listdir(log_dir):
                if not filename.startswith(base_name):
                    continue

                if filename == base_name:
                    continue

                filepath = os.path.join(log_dir, filename)

                if os.path.isfile(filepath):
                    file_mtime = datetime.fromtimestamp(os.path.getmtime(filepath))
                    if file_mtime < cutoff_date:
                        os.remove(filepath)
        except Exception as e:
            print(f"⚠️ Error during log cleanup: {e}")

    def _flush_buffer(self):
        if not self._buffer:
            return

        buffer_to_write = self._buffer
        self._buffer = []

        try:
            with open(self._filename, "a", encoding="utf-8") as f:
                f.writelines(buffer_to_write)
        except Exception as e:
            print(f"⚠️ Buffer write error: {e}")
            self._buffer = buffer_to_write + self._buffer

    def emit(self, record: LogRecord):
        """emit a log record to file."""
        if not self._should_log(record):
            return

        time_str = record.timestamp.strftime(self._time_format)
        context = record.context

        log_line = f"[{time_str}] | {record.level: <8} | {context} {record.message}\n"

        with self._lock:
            self._buffer.append(log_line)

            if len(self._buffer) >= self._buffer_size:
                self._flush_buffer()

            self._log_count += 1

            if self._log_count % self._check_rotation_every == 0:
                if self._should_rotate():
                    self._flush_buffer()
                    self._rotate_log()

    def close(self):
        """flush buffer and close the handler."""
        self._flush_buffer()



from .logger import logger, dLogger
from .handlers import Handler, ConsoleHandler, FileHandler, LogRecord, Filter
from .formatters import Formatter, SimpleFormatter, ExceptionFormatter
from .filters import LevelFilter, KeywordFilter, ModuleFilter

__version__ = "0.3.4"
__all__ = [
    "logger",
    "dLogger",
    "Handler",
    "ConsoleHandler",
    "FileHandler",
    "LogRecord",
    "Filter",
    "Formatter",
    "SimpleFormatter",
    "ExceptionFormatter",
    "LevelFilter",
    "KeywordFilter",
    "ModuleFilter",
]




import sys
import traceback
from typing import Optional

from ..handlers.base import Formatter

class ExceptionFormatter(Formatter):
    """formatter that formats exceptions with traceback."""

    def format(self, record: "handlers.LogRecord") -> str:
        """format a log record with exception traceback."""
        raise NotImplementedError

    @staticmethod
    def format_exception(exc: Optional[BaseException]) -> str:
        """
        format exception with full traceback.

        args:
            exc: exception object

        returns:
            formatted traceback string
        """
        if exc:
            return ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__))
        return ""

    @staticmethod
    def get_current_exception() -> Optional[BaseException]:
        """
        get current exception from sys.exc_info().

        returns:
            current exception or None
        """
        exc_info = sys.exc_info()
        return exc_info[1]




from ..handlers.base import Filter, LogRecord

class ModuleFilter(Filter):
    """filter that includes only records from specific modules."""

    def __init__(self, modules: list[str]):
        self.modules = modules

    def filter(self, record: LogRecord) -> bool:
        return any(mod in record.context for mod in self.modules)



from ..handlers.base import Filter
from .level import LevelFilter
from .keyword import KeywordFilter
from .module import ModuleFilter

__all__ = ["Filter", "LevelFilter", "KeywordFilter", "ModuleFilter"]



from .base import Handler, Formatter, LogRecord, Filter
from .console import ConsoleHandler
from .file import FileHandler

__all__ = ["Handler", "Formatter", "LogRecord", "Filter", "ConsoleHandler", "FileHandler"]




from ..handlers.base import Formatter, LogRecord

class SimpleFormatter(Formatter):
    """simple formatter that formats log records as plain text."""

    def __init__(self, time_format: str = "%Y-%m-%d %H:%M:%S"):
        self._time_format = time_format

    def format(self, record: LogRecord) -> str:
        """format a log record as plain text."""
        time_str = record.timestamp.strftime(self._time_format)
        return f"[{time_str}] | {record.level: <8} | {record.context} {record.message}"
